import unittest

from lalr.utils import Queue


class ParseTestCase(unittest.TestCase):
    def test_iteration(self):
        tasks = ['A', 'B', 'C', 'D', 'E', 'F']
        queue = Queue(tasks)

        self.assertEqual([task for task in queue], tasks)

    def test_insertion_during_iteration(self):
        queue = Queue([1])

        for task in queue:
            if task >= 128:
                break
            queue.add(task * 2)

        self.assertEqual(
            list(queue.processed),
            [1, 2, 4, 8, 16, 32, 64, 128],
        )

    def test_reinsertion_before_pop(self):
        tasks = [1, 2, 3, 4, 5]
        queue = Queue(tasks)

        processed = []
        processed += [queue.pop(), queue.pop()]
        queue.add(3)

        processed += [queue.pop(), queue.pop(), queue.pop()]

        self.assertEqual(processed, tasks)
        self.assertFalse(queue)
        self.assertRaises(IndexError, queue.pop)

    def test_reinsertion_after_pop(self):
        tasks = [1, 2, 3, 4, 5]
        queue = Queue(tasks)

        processed = []
        processed += [queue.pop(), queue.pop(), queue.pop()]
        queue.add(3)

        processed += [queue.pop(), queue.pop()]

        self.assertEqual(processed, tasks)
        self.assertFalse(queue)
        self.assertRaises(IndexError, queue.pop)

    def test_processed(self):
        queue = Queue([1, 2, 3, 4, 5])
        processed = set([queue.pop(), queue.pop()])
        self.assertEqual(set(queue.processed), processed)



import enum
import unittest

import lalr.exceptions
from lalr import Parser, Production, _parse_production_spec


class ProductionMiniLanguageTestCase(unittest.TestCase):
    def test_symbol(self):
        symbols, bindings = _parse_production_spec('a')

        self.assertEqual(symbols, ('a',))
        self.assertEqual(bindings, (None,))

    def test_symbol_binding(self):
        symbols, bindings = _parse_production_spec('symbol:binding')

        self.assertEqual(symbols, ('symbol',))
        self.assertEqual(bindings, ('binding',))

    def test_multiple_symbol_bindings(self):
        symbols, bindings = _parse_production_spec(
            'symbol_a:binding_a symbol_b:binding_b',
        )

        self.assertEqual(symbols, ('symbol_a', 'symbol_b'))
        self.assertEqual(bindings, ('binding_a', 'binding_b'))

    def test_unbound_symbols(self):
        symbols, bindings = _parse_production_spec(
            'symbol_a symbol_b:binding_b',
        )

        self.assertEqual(symbols, ('symbol_a', 'symbol_b'))
        self.assertEqual(bindings, (None, 'binding_b'))

    def test_whitespace(self):
        symbols, bindings = _parse_production_spec('''
            symbol_a: binding_a
            symbol_b : binding_b
        ''')

        self.assertEqual(symbols, ('symbol_a', 'symbol_b'))
        self.assertEqual(bindings, ('binding_a', 'binding_b'))


class ParseTestCase(unittest.TestCase):
    def test_example(self):
        parser = Parser([
            Production("N", ("V", "=", "E")),
            Production("N", ("E",)),
            Production("E", ("V",)),
            Production("V", ("x",)),
            Production("V", ("*", "E")),
        ], "N")

        parser.parse(["x", "=", "*", "x"])

    def test_bad_example(self):
        parser = Parser([
            Production("N", ("V", "=", "E")),
            Production("N", ("E",)),
            Production("E", ("V",)),
            Production("V", ("x",)),
            Production("V", ("*", "E")),
        ], "N")

        with self.assertRaises(lalr.exceptions.ParseError):
            parser.parse(["x", "*", "x"])

    def test_enum_terminals(self):
        class Terminal(enum.Enum):
            VAR = enum.auto()
            EQ = enum.auto()
            STAR = enum.auto()

        parser = Parser([
            Production("N", ("V", Terminal.EQ, "E")),
            Production("N", ("E",)),
            Production("E", ("V",)),
            Production("V", (Terminal.VAR,)),
            Production("V", (Terminal.STAR, "E")),
        ], "N")

        parser.parse([Terminal.VAR, Terminal.EQ, Terminal.STAR, Terminal.VAR])

    def test_enum_nonterminals(self):
        class NonTerminal(enum.Enum):
            N = enum.auto()
            E = enum.auto()
            V = enum.auto()

        parser = Parser([
            Production(NonTerminal.N, (NonTerminal.V, "=", NonTerminal.E)),
            Production(NonTerminal.N, (NonTerminal.E,)),
            Production(NonTerminal.E, (NonTerminal.V,)),
            Production(NonTerminal.V, ("x",)),
            Production(NonTerminal.V, ("*", NonTerminal.E)),
        ], NonTerminal.N)

        parser.parse(["x", "=", "*", "x"])



import re

from lalr.analysis import ParseTable
from lalr.exceptions import ProductionSpecParseError
from lalr.grammar import Grammar, InternalProduction
from lalr.parsing import parse


_default = object()


_production_spec_symbol_re = re.compile(r'''
    \s*
    (?P<symbol>
        [a-zA-Z0-9-_]+
    )
    (?:
        \ *
        :
        \ *
        (?P<binding>
            [a-zA-Z0-9-_]+
        )
    )?
    \s*
''', re.VERBOSE)  # pylint: disable=no-member


def _parse_production_spec(string):
    cursor = 0

    symbols = []
    bindings = []

    while cursor < len(string):
        mo = _production_spec_symbol_re.match(string, cursor)
        if mo is None:
            raise ProductionSpecParseError('Invalid production spec.')

        assert mo.group('symbol')

        symbols.append(mo.group('symbol'))
        bindings.append(mo.group('binding'))

        cursor = mo.end()

    return tuple(symbols), tuple(bindings)


def _default_action(**kwargs):
    return None


class Production(object):

    __slots__ = ('_name', '_symbols', '_bindings', '_action')

    def __init__(self, name, symbols, bindings=_default, action=None):
        self._name = name

        if isinstance(symbols, str):
            if bindings is not _default:
                raise TypeError(
                    'Production expects either a string spec or tuples of '
                    'symbols and bindings'
                )

            self._symbols, self._bindings = _parse_production_spec(symbols)
        else:
            self._symbols = tuple(symbols)

            if bindings is _default:
                bindings = tuple([None] * len(self._symbols))
            self._bindings = tuple(bindings)

        if action is None:
            action = _default_action
        self._action = action

    @property
    def name(self):
        return self._name

    @property
    def symbols(self):
        """The tuple of symbols that the production matches
        """
        return self._symbols

    @property
    def bindings(self):
        """A tuple of optional argument names that describe how parsed symbols
        should be passed to the action.
        """
        return self._bindings

    def action(self, *values, state=_default):
        kwargs = {
            binding: value
            for binding, value in zip(self.bindings, values)
            if binding is not None
        }

        if state is not _default:
            kwargs.update(state=state)

        return self._action(**kwargs)

    def __len__(self):
        return len(self.symbols)

    def __getitem__(self, index):
        return self.symbols[index]

    def __str__(self):
        return "{name} -> {symbols}".format(
            name=self.name,
            symbols=' '.join(str(symbol) for symbol in self.symbols),
        )

    def __repr__(self):
        return "[" + str(self) + "]"

    def __eq__(self, other):
        return (
            self.name == other.name and
            self.symbols == other.symbols and
            self.action == other.action
        )

    def __hash__(self):
        return (
            hash(self.name) ^
            hash(self.symbols) ^
            hash(self.action)
        )


class Parser(object):
    def __init__(self, productions, target):

        grammar_productions = []
        self._actions = {}

        for production in productions:
            grammar_production = InternalProduction(
                production.name, production.symbols,
            )
            self._actions[grammar_production] = production.action
            grammar_productions.append(grammar_production)

        grammar = Grammar(grammar_productions)

        self._parse_table = ParseTable(grammar, target)

    def parse(self, tokens, *, state=_default):
        def _action(production, *values):
            if state is not _default:
                return self._actions[production](*values, state=state)
            else:
                return self._actions[production](*values)

        return parse(
            self._parse_table, tokens, action=_action,
            token_symbol=self.token_symbol, token_value=self.token_symbol,
        )

    @staticmethod
    def token_symbol(token):
        return token

    @staticmethod
    def token_value(token):
        return token


def compile(productions, target):
    return Parser(productions, target)



class CompilationError(Exception):
    pass


class ProductionSpecParseError(CompilationError):
    pass


class ConflictError(CompilationError):
    pass


class ShiftReduceConflictError(ConflictError):
    pass


class ReduceReduceConflictError(ConflictError):
    pass


class ParseError(Exception):
    pass



from typing import Generic, Hashable, Iterable, List, Set, TypeVar


T = TypeVar('T', bound=Hashable)


class Queue(Iterable[T], Generic[T]):
    """
    A simple task queue that will yield the items added to it exactly than
    once.
    """
    def __init__(self, items: Iterable[T] = []) -> None:
        # The set of all items that have been added to the queue.
        # TODO:  Benchmark to determine if this is actually faster.
        self._by_identity: Set[T] = set()

        # A list, in order, of all items that have been added to the queue.
        self._by_order: List[T] = list()

        # The index of the next item to be processed in the `_by_order` array.
        self._cursor: int = 0

        self.update(items)

    def add(self, item: T):
        """
        Attempts to add an item to the end of the queue.

        If the item is already in the queue, or has already been processed,
        this call will have no effect.

        :param item:
            The item to add.
        """
        if item not in self._by_identity:
            self._by_order.append(item)
            self._by_identity.add(item)

    def update(self, items: Iterable[T]):
        """
        Adds all new items in iterable, in order, to the end of the queue.

        :param items:
            An iterable of items to process.
        """
        for item in items:
            self.add(item)

    def pop(self) -> T:
        """
        Removes and returns the next item that is waiting to be processed.

        Raises :exception:`IndexError` if there are no unprocessed items
        remaining.

        :return:
            The next item in the queue.
        """
        if self._cursor == len(self._by_order):
            raise IndexError("queue is empty")
        value = self._by_order[self._cursor]
        self._cursor += 1
        return value

    @property
    def processed(self) -> Iterable[T]:
        """
        Returns an iterable over all of the items in the queue that have
        already been processed.

        The items will be returned in their original order.

        :return:
            An iterable of work items.
        """
        return self._by_order[:self._cursor]

    def __len__(self) -> int:
        """
        Returns the number of unprocessed items in the queue.
        """
        return len(self._by_order) - self._cursor

    def __bool__(self) -> bool:
        """
        Returns True if there are any items that are waiting to be processed.
        """
        return self._cursor != len(self._by_order)

    def __iter__(self):
        return self

    def __next__(self) -> T:
        try:
            return self.pop()
        except IndexError:
            raise StopIteration()



from types import MappingProxyType

from lalr.utils import Queue


class InternalProduction(object):

    __slots__ = ('_name', '_symbols')

    def __init__(self, name, symbols, action=None):
        self._name = name
        self._symbols = tuple(symbols)

    @property
    def name(self):
        return self._name

    @property
    def symbols(self):
        return self._symbols

    def __len__(self):
        return len(self.symbols)

    def __getitem__(self, index):
        return self.symbols[index]

    def __str__(self):
        return "{name} -> {symbols}".format(
            name=self.name,
            symbols=' '.join(str(symbol) for symbol in self.symbols),
        )

    def __repr__(self):
        return "[" + str(self) + "]"

    def __eq__(self, other):
        return (
            self.name == other.name and
            self.symbols == other.symbols
        )

    def __hash__(self):
        return (
            hash(self.name) ^
            hash(self.symbols)
        )


class Grammar(object):

    def __init__(self, productions):
        self._productions = frozenset(productions)

        # Force first_sets property to evaluate
        self.first_sets

        # There must be a first set for every symbol in the grammar
        assert (
            set(self.first_sets) ==
            set.union(self.terminals, self.nonterminals)
        )

        # First sets should contain only terminals
        assert not any(
            set.difference(first_set, self.terminals)
            for first_set in self.first_sets.values()
        )

        # There should be a first set for every terminal and non-terminal
        assert self.symbols == set(self.first_sets.keys())

    @property
    def terminals(self):
        """
        The set of all symbols that appear on the right hand side of one or
        more productions but have no production of their own.
        """
        return set.difference(self.symbols, self.nonterminals)

    @property
    def nonterminals(self):
        """
        The set of all symbols for which there are production rules.
        """
        return {production.name for production in self.productions()}

    @property
    def symbols(self):
        """
        The set of all symbols, both terminal and non-terminal.
        """
        symbols = set()
        for production in self.productions():
            symbols.add(production.name)
            symbols.update(production.symbols)
        return symbols

    def is_terminal(self, symbol):
        return symbol in self.terminals

    def is_nonterminal(self, symbol):
        return symbol in self.nonterminals

    def productions(self, name=None):
        if name is None:
            return set(self._productions)

        return {
            production for production in self._productions
            if production.name == name
        }

    def _build_first_sets(self):
        # A map from symbols to sets of symbols for which there exist
        # productions where the first symbol is the first element
        has_first_symbol = {}

        for production in self.productions():
            has_first_symbol.setdefault(
                production.symbols[0], set()
            ).add(production.name)

        # A map from symbols to the set of terminals which can appear as the
        # first terminal in a string that the symbol matches
        first_sets = {}

        for terminal in self.terminals:
            # TODO if this is needed it should be folded into the inner loop.
            # kept separate for now so that it's very clear what is happening
            first_sets[terminal] = {terminal}

            if terminal not in has_first_symbol:
                continue

            queue = Queue(has_first_symbol[terminal])

            while queue:
                nonterminal = queue.pop()
                first_sets.setdefault(nonterminal, set()).add(terminal)

                queue.update(has_first_symbol.get(nonterminal, set()))

        return first_sets

    @property
    def first_sets(self):
        """
        A map from symbols to sets of terminals that can appear as the first
        terminal in that symbol.  Terminal symbols are included and just map to
        a set containing only themselves
        """
        if not hasattr(self, '_first_sets'):
            self._first_sets = MappingProxyType(self._build_first_sets())

        return self._first_sets



import unittest

from lalr.tests import test_analysis, test_grammar, test_parsing, test_queue


loader = unittest.TestLoader()
suite = unittest.TestSuite((
    loader.loadTestsFromModule(test_queue),
    loader.loadTestsFromModule(test_parsing),
    loader.loadTestsFromModule(test_analysis),
    loader.loadTestsFromModule(test_grammar),
))



import unittest

from lalr.analysis import _build_item_set, _build_transition_table, _Item
from lalr.grammar import Grammar, InternalProduction


class _ItemSetTestCase(unittest.TestCase):
    def test_zero(self):
        grammar = Grammar([])

        starting_item = _Item(
            InternalProduction('S', ('a',)), 0, {'$'},
        )

        item_set = _build_item_set(grammar, {starting_item})
        self.assertEqual(item_set.items, {starting_item})

    def test_one(self):
        grammar = Grammar([
            InternalProduction('A', ('a',)),
        ])

        starting_item = _Item(
            InternalProduction('S', ('A',)), 0, {'$'},
        )

        item_set = _build_item_set(grammar, {starting_item})

        self.assertEqual(item_set.items, {
            _Item(InternalProduction('S', ('A',)), 0, {'$'}),
            _Item(InternalProduction('A', ('a',)), 0, {'$'}),
        })

    def test_example(self):
        grammar = Grammar([
            InternalProduction("N", ("V", "=", "E")),
            InternalProduction("N", ("E",)),
            InternalProduction("E", ("V",)),
            InternalProduction("V", ("x",)),
            InternalProduction("V", ("*", "E")),
        ])

        sets, transitions = _build_transition_table(grammar, "N")

        for num, item_set in enumerate(sets):
            print(num)
            for item in item_set.kernel:
                print(item)
            for item in item_set.derived:
                print('+ ' + str(item))
            print()

            print(transitions[num])



import unittest

from lalr.grammar import Grammar, InternalProduction


class TerminalsTestCase(unittest.TestCase):
    def test_loop(self):
        grammar = Grammar([
            InternalProduction("A", ("B",)),
            InternalProduction("A", ("x",)),
            InternalProduction("B", ("A",)),
        ], )
        self.assertEqual(grammar.terminals, {"x"})

    def test_example(self):
        grammar = Grammar([
            InternalProduction("N", ("V", "=", "E")),
            InternalProduction("N", ("E",)),
            InternalProduction("E", ("V",)),
            InternalProduction("V", ("x",)),
            InternalProduction("V", ("*", "E")),
        ])
        self.assertEqual(grammar.terminals, {"x", "=", "*"})


class FirstSetsTestCase(unittest.TestCase):
    def test_loop(self):
        grammar = Grammar([
            InternalProduction("A", ("B",)),
            InternalProduction("A", ("x",)),
            InternalProduction("B", ("A",)),
        ])

        self.assertEqual(grammar.first_sets, {
            "x": {"x"},
            "A": {"x"},
            "B": {"x"},
        })

    def test_example(self):
        grammar = Grammar([
            InternalProduction("N", ("V", "=", "E")),
            InternalProduction("N", ("E",)),
            InternalProduction("E", ("V",)),
            InternalProduction("V", ("x",)),
            InternalProduction("V", ("*", "E")),
        ])

        self.assertEqual(grammar.first_sets, {
            "x": {"x"},
            "=": {"="},
            "*": {"*"},
            "N": {"*", "x"},
            "E": {"*", "x"},
            "V": {"*", "x"},
        })



from setuptools import find_packages, setup


setup(
    name='lalr',
    url='https://github.com/bwhmather/python-lalr',
    version='0.0.1',
    author='Ben Mather',
    author_email='bwhmather@bwhmather.com',
    maintainer='Ben Mather',
    license='BSD',
    description=(
        "A pure python LALR parser generator"
    ),
    long_description=__doc__,
    classifiers=[
        'Development Status :: 1 - Planning',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3 :: Only',
    ],
    install_requires=[],
    packages=find_packages(),
    package_data={
        '': ['*.*'],
    },
    test_suite='lalr.tests.suite',
)
